/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package fr.amap.lidar.amapvox.voxelisation.postproc;

import fr.amap.commons.util.io.file.FileManager;
import fr.amap.commons.util.TimeCounter;
import fr.amap.lidar.amapvox.commons.VoxelSpaceInfos;
import fr.amap.lidar.amapvox.commons.GTheta;
import fr.amap.lidar.amapvox.commons.LADParams;
import fr.amap.lidar.amapvox.commons.LeafAngleDistribution;
import fr.amap.lidar.amapvox.commons.Voxel;
import fr.amap.lidar.amapvox.voxelisation.configuration.VoxMergingCfg;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import org.apache.log4j.Logger;
import fr.amap.lidar.amapvox.util.AVoxTask;
import fr.amap.lidar.amapvox.voxelisation.Voxelization;
import java.io.File;
import java.math.BigInteger;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.HashMap;
import java.util.Locale;

/**
 * Voxel files merging.
 *
 * @author Julien Heurtebize (julienhtbe@gmail.com) 2016
 * @author Philippe Verley (philippe.verley@ird.fr) 2019
 */
public class VoxelFileMerging extends AVoxTask {

    private final static Logger LOGGER = Logger.getLogger(VoxelFileMerging.class);

    private VoxMergingCfg cfg;

    public VoxelFileMerging(File file, int ncpu) {
        super(file, ncpu);
    }

    public VoxelFileMerging(VoxMergingCfg cfg, int ncpu) {
        super(null, ncpu);
        this.cfg = cfg;
    }

    @Override
    public String getName() {
        return "Voxel file merging";
    }

    @Override
    public void init() throws Exception {

        if (null != file) {
            cfg = new VoxMergingCfg();
            cfg.readConfiguration(file);
        }

        setCancelled(false);
    }

    @Override
    public File[] call() throws Exception {

        int nFile = cfg.getFiles().size();

        // special case: no file to merge
        if (nFile <= 0) {
            LOGGER.info("No file to merge");
            return null;
        }

        // time merging process
        long startTime = System.currentTimeMillis();

        // variable declaration
        HashMap<Voxel.OutputVariable, Integer> column;
        HashMap<Integer, Voxel.OutputVariable> variables;
        int nRow;
        VoxelSpaceInfos infos = new VoxelSpaceInfos();
        double[][] mergedArray;
        int nColumn;

        // read meta-data from voxel file
        infos.readFromVoxelFile(cfg.getFiles().get(0));
        // LAD parameters
        LADParams ladParams = new LADParams();
        ladParams.setLadType(infos.getLadType());
        if (infos.getLadParams() != null) {
            ladParams.setLadBetaFunctionAlphaParameter((float) infos.getLadParams()[0]);
            ladParams.setLadBetaFunctionBetaParameter((float) infos.getLadParams()[1]);
        }
        cfg.getVoxelParameters().setLadParams(ladParams);

        nRow = infos.getSplit().x * infos.getSplit().y * infos.getSplit().z;
        nColumn = infos.getColumnNamesList().size();
        mergedArray = new double[nRow][nColumn];
        // map output variables and column index
        column = new HashMap(nColumn);
        variables = new HashMap();
        for (int iColumn = 0; iColumn < nColumn; iColumn++) {
            String name = infos.getColumnNamesList().get(iColumn);
            Voxel.OutputVariable variable = Voxel.OutputVariable.find(name);
            column.put(variable, iColumn);
            variables.put(iColumn, variable);
        }

        BigInteger[] subsampling = new BigInteger[nRow];
        int nsubvoxel = -1;
        if (column.containsKey(Voxel.OutputVariable.SUBSAMPLING)) {
            for (int iRow = 0; iRow < nRow; iRow++) {
                subsampling[iRow] = BigInteger.ZERO;
            }
            try {
                nsubvoxel = Integer.valueOf(infos.getProperty("nsubvoxel"));
            } catch (Exception ex) {
                LOGGER.warn("Missing parameter nsubvoxel in vox file header, exploration rate cannot be computed.");
                nanify(mergedArray, column.get(Voxel.OutputVariable.EXPLORATION_RATE));
            }
        }

        // loop over the voxel files
        for (int iFile = 0; iFile < nFile; iFile++) {

            // task cancelled by user
            if (isCancelled()) {
                return null;
            }

            // print progress
            String msg = "Merging file " + (iFile + 1) + "/" + nFile;
            LOGGER.info(msg);
            fireProgress(msg, (iFile + 1), nFile);

            // open ith vox file
            try (BufferedReader reader = new BufferedReader(new FileReader(cfg.getFiles().get(iFile)))) {

                // skip headers
                FileManager.skipLines(reader, infos.getNLine());

                int iRow = 0;
                String line;
                // loop over the rows
                while ((line = reader.readLine()) != null) {

                    String[] splitLine = line.split(" ");

                    if (splitLine.length != nColumn) {
                        LOGGER.error("Voxel file merging, number of columns mismatch");
                        return null;
                    }

                    // loop over the columns
                    for (int iColumn = 0; iColumn < nColumn; iColumn++) {
                        switch (variables.get(iColumn).getMode()) {
                            case SUM:
                                mergedArray[iRow][iColumn] += Double.valueOf(splitLine[iColumn]);
                                break;
                            case NO_MERGE:
                                if (iFile == 0) {
                                    mergedArray[iRow][iColumn] = Double.valueOf(splitLine[iColumn]);
                                }
                                break;
                            case NAN:
                                if (iFile == 0) {
                                    mergedArray[iRow][iColumn] = Double.NaN;
                                }
                                break;
                            case WEIGHTED_MEAN:
                                // weigthed mean sum(variable * nSampling) / sum(nSampling)
                                // 1st part: sum(variable * nSampling)
                                if (column.containsKey(Voxel.OutputVariable.NUMBER_OF_SHOTS)) {
                                    int nSampling = Integer.valueOf(splitLine[column.get(Voxel.OutputVariable.NUMBER_OF_SHOTS)]);
                                    if (nSampling > 0) {
                                        mergedArray[iRow][iColumn] += nSampling * Double.valueOf(splitLine[iColumn]); // mergedArray initialized with zeros
                                    }
                                }
                                break;
                            case CUSTOM:
                                // handled separately further down
                                break;
                            case REMOVE:
                                // nothing to do
                                break;
                        }
                    }

                    // numerical estimation of the transmittance
                    // weighted mean sum(trNumEstim * bsEntering) / sum(bsEntering)
                    // 1st part: sum(trNumEstim * bsEntering)
                    if (column.containsKey(Voxel.OutputVariable.ESTIMATED_TRANSMITTANCE)
                            && column.containsKey(Voxel.OutputVariable.ENTERING_BEAM_SURFACE)) {
                        int iEstimTr = column.get(Voxel.OutputVariable.ESTIMATED_TRANSMITTANCE);
                        double trNumEstim = Double.valueOf(splitLine[iEstimTr]);
                        if (!Double.isNaN(trNumEstim)) {
                            double bsEntering = Double.valueOf(splitLine[column.get(Voxel.OutputVariable.ENTERING_BEAM_SURFACE)]);
                            mergedArray[iRow][iEstimTr] += bsEntering * trNumEstim; // mergedArray initialized with zeros
                        }
                    }

                    // exploration rate
                    // step 1 merge subSampling variable
                    if (nsubvoxel > 0 && column.containsKey(Voxel.OutputVariable.SUBSAMPLING)) {
                        String subSamplingRow = splitLine[column.get(Voxel.OutputVariable.SUBSAMPLING)];
                        if (isBigInteger(subSamplingRow)) {
                            subsampling[iRow] = subsampling[iRow].or(new BigInteger(subSamplingRow, 36));
                        }
                    }

                    // move to next row
                    iRow++;
                } // end loop over the rows
            } // closes current voxel file
        } // end loop over the voxel files

        // weighted mean variables
        // weigthed mean sum(variable * nSampling) / sum(nSampling)
        // 2nd part: division by sum(nSampling)
        for (int iColumn = 0; iColumn < nColumn; iColumn++) {
            if (Voxel.MergeMode.WEIGHTED_MEAN == variables.get(iColumn).getMode()) {
                if (column.containsKey(Voxel.OutputVariable.NUMBER_OF_SHOTS)) {
                    // column index
                    int iSampling = column.get(Voxel.OutputVariable.NUMBER_OF_SHOTS);
                    //  weigthed mean of sum(variable * nSampling) / sum(nSampling)
                    for (int iRow = 0; iRow < nRow; iRow++) {
                        mergedArray[iRow][iColumn] = mergedArray[iRow][iSampling] > 0
                                ? mergedArray[iRow][iColumn] / mergedArray[iRow][iSampling]
                                : Double.NaN;  // merged voxel not sampled
                    }
                } else {
                    LOGGER.warn("Cannot merge " + variables.get(iColumn).getShortName() + ", nbSampling column is missing for weighted mean.");
                    nanify(mergedArray, iColumn);
                }
            }
        }

        // mean ray length
        // sum(totalLength) / sum(nSampling)
        boolean lMeanTotalMerged = false;
        if (column.containsKey(Voxel.OutputVariable.MEAN_TOTAL_LENGTH)) {

            if (column.containsKey(Voxel.OutputVariable.TOTAL_LENGTH)
                    && column.containsKey(Voxel.OutputVariable.NUMBER_OF_SHOTS)) {
                // column index
                int iTotalLength = column.get(Voxel.OutputVariable.TOTAL_LENGTH);
                int iSampling = column.get(Voxel.OutputVariable.NUMBER_OF_SHOTS);
                int iMeanLength = column.get(Voxel.OutputVariable.MEAN_TOTAL_LENGTH);
                // loop over rows
                for (int iRow = 0; iRow < nRow; iRow++) {
                    mergedArray[iRow][iMeanLength] = mergedArray[iRow][iSampling] > 0
                            ? mergedArray[iRow][iTotalLength] / mergedArray[iRow][iSampling]
                            : Double.NaN; // merged voxel not sampled
                }
                lMeanTotalMerged = true;
            } else {
                LOGGER.warn("Cannot merge lMeanTotal: nbSampling or lgTotal columns are missing.");
                nanify(mergedArray, column.get(Voxel.OutputVariable.DIST_LASER));
            }
        }

        // approximated transmittance
        boolean transmittanceMerged = false;
        if (column.containsKey(Voxel.OutputVariable.APPROXIMATED_TRANSMITTANCE)) {
            if (column.containsKey(Voxel.OutputVariable.ENTERING_BEAM_SURFACE)
                    && column.containsKey(Voxel.OutputVariable.INTERCEPTED_BEAM_SURFACE)
                    && lMeanTotalMerged) {
                // column index
                int iMeanLength = column.get(Voxel.OutputVariable.MEAN_TOTAL_LENGTH);
                int iBsEnt = column.get(Voxel.OutputVariable.ENTERING_BEAM_SURFACE);
                int iBsInterp = column.get(Voxel.OutputVariable.INTERCEPTED_BEAM_SURFACE);
                int iTransm = column.get(Voxel.OutputVariable.APPROXIMATED_TRANSMITTANCE);
                // loop over rows
                for (int iRow = 0; iRow < nRow; iRow++) {
                    mergedArray[iRow][iTransm] = Voxelization.approximateTransmittance(mergedArray[iRow][iBsEnt], mergedArray[iRow][iBsInterp], mergedArray[iRow][iMeanLength]);
                }
                transmittanceMerged = true;
            } else {
                LOGGER.warn("Cannot merge transmittance: bsEntering, bsIntercepted, lgTotal and/or nbSampling columns are missing.");
                nanify(mergedArray, column.get(Voxel.OutputVariable.APPROXIMATED_TRANSMITTANCE));
            }
        }

        // plant area density
        if (column.containsKey(Voxel.OutputVariable.PLANT_AREA_DENSITY)) {
            boolean meanAngleMerged = column.containsKey(Voxel.OutputVariable.MEAN_ANGLE) && column.containsKey(Voxel.OutputVariable.NUMBER_OF_SHOTS);
            if (transmittanceMerged && meanAngleMerged) {
                // column index
                int iTransm = column.get(Voxel.OutputVariable.APPROXIMATED_TRANSMITTANCE);
                int iPAD = column.get(Voxel.OutputVariable.PLANT_AREA_DENSITY);
                int iAngle = column.get(Voxel.OutputVariable.MEAN_ANGLE);
                // leaf angle distribution
                LADParams ladParameters = cfg.getVoxelParameters().getLadParams();
                if (ladParameters == null) {
                    ladParameters = new LADParams();
                }
                LeafAngleDistribution distribution = new LeafAngleDistribution(ladParameters.getLadType(),
                        ladParameters.getLadBetaFunctionAlphaParameter(),
                        ladParameters.getLadBetaFunctionBetaParameter());
                GTheta direcTransmittance = new GTheta(distribution);
                direcTransmittance.buildTable(GTheta.DEFAULT_STEP_NUMBER);
                // loop over rows
                for (int iRow = 0; iRow < nRow; iRow++) {
                    mergedArray[iRow][iPAD] = Voxelization.computePADFromNormTransmittance(mergedArray[iRow][iTransm], mergedArray[iRow][iAngle], infos.getMaxPAD(), direcTransmittance);
                }
            } else {
                LOGGER.warn("Cannot merge PAD: transmittance or angleMean columns are missing.");
                nanify(mergedArray, column.get(Voxel.OutputVariable.PLANT_AREA_DENSITY));
            }
        }

        // numerical estimation of the transmittance
        // weighted mean sum(trNumEstim * bsEntering) / sum(bsEntering)
        // 2nd part: division by sum(bsEntering)
        if (column.containsKey(Voxel.OutputVariable.ESTIMATED_TRANSMITTANCE)) {
            if (column.containsKey(Voxel.OutputVariable.ENTERING_BEAM_SURFACE)) {
                // column index
                int iTrNum = column.get(Voxel.OutputVariable.ESTIMATED_TRANSMITTANCE);
                int iBsEnt = column.get(Voxel.OutputVariable.ENTERING_BEAM_SURFACE);
                // loop over rows
                for (int iRow = 0; iRow < nRow; iRow++) {
                    // trNumEstim as weighted mean of sum(trNumEstim * bsEntering) / sum(bsEntering)
                    mergedArray[iRow][iTrNum] = mergedArray[iRow][iBsEnt] > 0
                            ? mergedArray[iRow][iTrNum] / mergedArray[iRow][iBsEnt]
                            : Double.NaN;  // merged voxel not sampled
                }
            } else {
                LOGGER.warn("Cannot merge trNumEstim: bsEntering column is missing.");
                nanify(mergedArray, column.get(Voxel.OutputVariable.ESTIMATED_TRANSMITTANCE));
            }
        }

        // exploration rate
        // step 2 compute merged exploration rate, number of explored sub voxels 
        // divided by number of sub voxels
        if (nsubvoxel > 0 && column.containsKey(Voxel.OutputVariable.SUBSAMPLING)) {
            if (column.containsKey(Voxel.OutputVariable.NUMBER_OF_SHOTS)) {
                int iExplo = column.get(Voxel.OutputVariable.EXPLORATION_RATE);
                int iSampling = column.get(Voxel.OutputVariable.NUMBER_OF_SHOTS);
                for (int iRow = 0; iRow < nRow; iRow++) {
                    mergedArray[iRow][iExplo] = mergedArray[iRow][iSampling] > 0
                            ? subsampling[iRow].bitCount() / (double) nsubvoxel
                            : Double.NaN;
                }
            } else {
                LOGGER.warn("Cannot merge explorationRate: nbSampling column is missing.");
                nanify(mergedArray, column.get(Voxel.OutputVariable.EXPLORATION_RATE));
            }
        }

        LOGGER.info("writing output file: " + cfg.getOutputFile().getAbsolutePath());
        long start_time = System.currentTimeMillis();

        try (BufferedWriter writer = new BufferedWriter(new FileWriter(cfg.getOutputFile()))) {
            // write header
            writer.write(infos.headerToString() + "\n");

            // write column names
            StringBuilder columnNames = new StringBuilder();
            // write header
            for (int iColumn = 0; iColumn < nColumn; iColumn++) {
                Voxel.OutputVariable variable = variables.get(iColumn);
                if (Voxel.MergeMode.REMOVE != variable.getMode()) {
                    columnNames.append(variable.getShortName()).append(" ");
                }
            }
            writer.write(columnNames.toString().trim() + "\n");

            DecimalFormat decimalFormat = (DecimalFormat) NumberFormat.getNumberInstance(Locale.US);
            decimalFormat.applyPattern("#0.############");
            decimalFormat.setGroupingUsed(false);
            decimalFormat.setMaximumFractionDigits(infos.getFractionDigits());

            for (int iRow = 0; iRow < nRow; iRow++) {
                StringBuilder voxel = new StringBuilder();
                for (int iColumn = 0; iColumn < nColumn; iColumn++) {
                    Voxel.OutputVariable variable = variables.get(iColumn);
                    if (Voxel.MergeMode.REMOVE != variable.getMode()) {
                        switch (variable.getType()) {
                            case INT:
                                voxel.append((int) mergedArray[iRow][iColumn]);
                                break;
                            default:
                                voxel.append(Double.isNaN(mergedArray[iRow][iColumn]) ? Double.NaN : decimalFormat.format(mergedArray[iRow][iColumn]));
                        }
                        voxel.append(" ");
                    }
                }
                writer.write(voxel.toString().trim() + "\n");
            }

            LOGGER.info("file written ( " + TimeCounter.getElapsedStringTimeInSeconds(start_time) + " )");

        } catch (IOException ex) {
            LOGGER.error(ex);
        }

        fireFinished(TimeCounter.getElapsedTimeInSeconds(startTime));

        return new File[]{cfg.getOutputFile()};
    }

    /**
     * Fill specified column of the array with NaN.
     *
     * @param array
     * @param iColumn
     */
    private void nanify(double[][] array, int iColumn) {
        for (int k = 0; k < array.length; k++) {
            array[k][iColumn] = Double.NaN;
        }
    }

    /**
     * Return true if the string is a valid binary representation.
     *
     * @param str
     * @return
     */
    private boolean isBigInteger(String str) {
        return (null != str) && (!str.isEmpty()) && str.matches("^[0-9a-z]*$");
    }
}
